/*
  This file is part of the ChipChop.io Arduino library.
  Copyright (c) 2024 ChipChop.io. All rights reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/



#ifndef CHIPCHOP_H
#define CHIPCHOP_H

 #include <Arduino.h>


class ChipChopManager{

    using commandCallbackType = void (*)(String,String,String,int);
    using errorCallbackType = void (*)(String);
   
 
    private:

        String  _SERVER = "";
        int _TCP_PORT = 8883;

        void findTCPPort();

        const String _quote = "\"";

        typedef struct{
            String event;
            String status;
        } _EVENT;

        bool _CAN_RUN = 1;

        _EVENT _send_buffer[11];
        int buffer_size = 0;
        unsigned long _send_rate = 510;
        unsigned long _send_pulse_timer = 0;
        void sendNextEvent();
        void addEventToBuffer(String type, String status);
        void popBufferQueue(int poz);
        void limitPopBufferQueue(String event, String status);
        void clearBufferQueue();

        bool _LOG_ENABLED = false;
        bool _LOG_RESPONSE_ONLY = false;

        String _UUID;
        String _DEVICE_ID;
        String _AUTH_CODE;


        unsigned long lastPollTime = 0;
       
        unsigned long _HEARTBEAT_INTERVAL = 10000; // setting to 0 changes the device to a "status request" only mode
        unsigned long lastHeartbeatSendTime = 0;
        unsigned long lastDateTimeRefresh = 0;
        bool _can_send_heartbeat = 1;

        unsigned long eventSpacing = 0;
        unsigned long lastEventTime = 0;
        byte eventsSent = 0;

        unsigned long _SERVER_TIME = 0; // received server time in millis used to run a real time clock
        unsigned long _DEVICE_TIME_MILLIS = 0;
        
       
       void _respond_to_status_request(const String& message_id);
       void _302_api_redirect(String api_uri);

        bool socket_restart_mode = 0;
        unsigned long socket_restart_timer = 0;
        int socket_restart_timeout = 0;

        void _send_heartbeat();
        void _validate_trigger_event(String component_name);
        void _send_trigger_event(String status);

        void _connect_socket();

        String zeroPrefix(int num);

        void setComponentStatus(const String component, String value);
        String componentsToJson(String triggerEventComponent);

        unsigned long stringToLong(String input);
        float stringToFloat(String input);

        commandCallbackType commandReceivedCallback;
        bool hasCommandCallback = false;
    

        void sendCommandCallback(String target, String value, String source, int command_age){
            if(hasCommandCallback == true){
                commandReceivedCallback(target, value, source, command_age);
            }
        }

        errorCallbackType errorReceivedCallback;
        bool hasErrorCallback = false;
                
        void sendErrorCallback(String error){
            if(hasErrorCallback == true){
                errorReceivedCallback(error);
            }
        }

        

        void setTime();
        bool _use_clock = false;

        ///The basic MD5 functions
        typedef unsigned long MD5_u32plus;

        typedef struct {
            MD5_u32plus lo, hi;
            MD5_u32plus a, b, c, d;
            unsigned char buffer[64];
            MD5_u32plus block[16];
        } MD5_CTX;

        static unsigned char* MD5_make_hash(char *arg);
        static unsigned char* MD5_make_hash(char *arg,size_t size);
        static char* MD5_make_digest(const unsigned char *digest, int len);
        static const void *MD5_body(void *ctxBuf, const void *data, size_t size);
        static void MD5Init(void *ctxBuf);
        static void MD5Final(unsigned char *result, void *ctxBuf);
        static void MD5Update(void *ctxBuf, const void *data, size_t size);

    public:

        typedef struct{
            String key;
            String value;
        } _JSON;
        int _MAX_COMPONENTS = 10;
        _JSON _COMPONENTS[10];

        void _process_server_message(const String& data);
        void connectionClosed();
        void restartSockets(){
            Serial.println(F("*** This function is deprecated"));
        }

        void pause();
        void unPause();

        unsigned long lastReconnectTime = 0;
        bool connected = false;
        bool handshakeComplete = 0;

        ChipChopManager();
        void run();

        void start(String server_uri, String uuid, String device_id, String auth_code );
        void closeConnection();
        void debug(bool enabled){
            _LOG_ENABLED = enabled;
        }
        void debug_resp_only(bool enabled){
            _LOG_ENABLED = enabled;
            _LOG_RESPONSE_ONLY = enabled;
        }

        void requestRate(int rate){
            if(rate < 500){
                rate = 500;
            }
            _send_rate = rate;
        }

        void log(const String &data);
        void log(const __FlashStringHelper *ifsh){ 
            log(reinterpret_cast<const char *>(ifsh)); 
        }

        void commandCallback(commandCallbackType cb){
            commandReceivedCallback = cb;
            hasCommandCallback = true;
        }
        void errorCallback(errorCallbackType cb){
            errorReceivedCallback = cb;
            hasErrorCallback = true;
        }

        String JSON_find(const String& obj, const String& property);
        String JSON_stringify(_JSON obj[], int obj_length);
        void toRGB(const String& obj, int (&rgb)[3]);

        // set the heartbeat interval in seconds, the minimum interval is 10 seconds and anything faster
        // will incur system violations marked on the account and potential dropping of the communication channel
        void heartBeatInterval(int interval){
            if(interval < 10){
                interval = 10;
            }
            _HEARTBEAT_INTERVAL = interval * 1000; //convert to milliseconds
        }

        void triggerEvent(String component_name,String status);
        void triggerEvent(String component_name,int status);
        void triggerEvent(String component_name,float status);

        void updateStatus(String component_name,String status);
        void updateStatus(String component_name,int status);
        void updateStatus(String component_name,float status);

        void removeStatus(String component_name);

        struct time_struct{
            int year;
            int yearShort;
            int month;
            int day;
            int hour;
            int minute;
            int second;
            unsigned long timestamp; //UNIX timestamp in seconds
            String monthLong;
            String monthShort;
            String formatted;
        };

        time_struct Date;
        String dateFormat = "dd MM YY - hh:mm:ss";
        void setDateFormat(String format){
            dateFormat = format;
        };

        void useClock(bool use){
            _use_clock = use;
        }

};

extern ChipChopManager ChipChop;

#endif



